import os
from pathlib import Path
from typing import List

from IPython import get_ipython
from IPython.core.magic import register_line_magic
from intake import Catalog

from framania.hvmania.manager import HVManiaManager
from framania.intakemania.extension import get_version_name, FramaniaExtendedIntakeSource
from framania.intakemania.extension import FramaniaExtendedIntakeCatalog


@register_line_magic
def start_nb(line):
    parts = line.split(' ')
    if len(parts) != 3:
        print('Make sure input `%start_nb {project title} {catalog file path} {data directory path}`')
        return

    analysis_name, catalog_file, data_dir = parts

    get_ipython().user_global_ns['__framania_pj_name__'] = analysis_name
    get_ipython().user_global_ns['__framania_catalog_file__'] = Path(catalog_file)
    get_ipython().user_global_ns['__framania_data_dir__'] = Path(data_dir)

    print(f'Start analysis notebook {analysis_name} (data directory: {Path(data_dir)})')


def _add_analysis_upstream(catalog, upstream_names: List[str]):
    upstreams = []
    for upstream_name in upstream_names:
        try:
            upstream = catalog.find_by_version_name(upstream_name)
        except:
            upstream = catalog.find_latest_source_by_name(upstream_name)
        upstreams.append(upstream)
    get_ipython().user_global_ns['__framania_analysis_upstream_names__'].extend(upstream_names)
    get_ipython().user_global_ns['__framania_analysis_upstreams__'].extend(upstreams)

    for upstream in upstreams:
        print(f'Register upstream as a global variable {upstream.name} from {upstream.name} {upstream.version}.')
        get_ipython().user_global_ns[upstream.name] = upstream.intake_source.to_dask()


@register_line_magic
def begin_analysis(line):
    parts = line.split(' ')
    if len(parts) < 2:
        print(
            'Make sure input `%begin_analysis {analysis name} {analysis version} {upstream analysis name1} {upstream analysis name2} ...`')
        return
    name, version = parts[:2]
    upstream_names = parts[2:]
    get_ipython().user_global_ns['__framania_analysis_name__'] = name
    get_ipython().user_global_ns['__framania_analysis_version__'] = version
    get_ipython().user_global_ns['__framania_analysis_upstreams__'] = []
    get_ipython().user_global_ns['__framania_analysis_upstream_names__'] = []

    proj_name = get_ipython().user_global_ns['__framania_pj_name__']
    print(f'Begin analysis {name} in {proj_name}')
    catalog = FramaniaExtendedIntakeCatalog(get_ipython().user_global_ns['__framania_catalog_file__'])
    _add_analysis_upstream(catalog, upstream_names)


@register_line_magic
def add_analysis_upstream(line):
    if len(line) == 0:
        print('Make sure input `%add_analysis_upstream {upstream analysis name1} {upstream analysis name2} ...`')
        return
    upstream_names = line.split(' ')
    catalog = FramaniaExtendedIntakeCatalog(get_ipython().user_global_ns['__framania_catalog_file__'])
    _add_analysis_upstream(catalog, upstream_names)


@register_line_magic
def end_analysis(line):
    if '__framania_analysis_name__' not in get_ipython().user_global_ns.keys():
        print('Make sure begin analysis before end analysis.')
        return

    if len(line) >= 1:
        result_var = get_ipython().user_global_ns[line]
        catalog = FramaniaExtendedIntakeCatalog(get_ipython().user_global_ns['__framania_catalog_file__'])
        name = get_ipython().user_global_ns['__framania_analysis_name__']
        version = get_ipython().user_global_ns['__framania_analysis_version__']
        upstreams = get_ipython().user_global_ns['__framania_analysis_upstreams__']
        data_dir = get_ipython().user_global_ns['__framania_data_dir__']
        print(f'Dump dask dataframe {line} to catalog as {name}')
        catalog.dump_dask(result_var, name, version, data_dir, upstreams)

    del get_ipython().user_global_ns['__framania_analysis_name__']
    del get_ipython().user_global_ns['__framania_analysis_version__']
    del get_ipython().user_global_ns['__framania_analysis_upstream_names__']
    del get_ipython().user_global_ns['__framania_analysis_upstreams__']

    print('End analysis.')


@register_line_magic
def serve_chart(line):
    parts = line.split(' ')
    if len(parts) < 3:
        print(
            'Make sure input `%serve_chart {chart name} {chart version} {chart variable}`')
        return

    pj_name = get_ipython().user_global_ns['__framania_pj_name__']
    name = get_ipython().user_global_ns['__framania_analysis_name__']
    version = get_ipython().user_global_ns['__framania_analysis_version__']
    data_dir = get_ipython().user_global_ns['__framania_data_dir__']

    chart_name, chart_version, result_var_name = parts
    result_var = get_ipython().user_global_ns[result_var_name]

    chart_dir = data_dir / 'chart' / get_version_name(name, version)

    manager = HVManiaManager(pj_name, chart_dir)
    manager[chart_name] = result_var

    catalog = FramaniaExtendedIntakeCatalog(get_ipython().user_global_ns['__framania_catalog_file__'])
    source = FramaniaExtendedIntakeSource(
        Catalog(),
        chart_name, chart_version, None,
        get_ipython().user_global_ns['__framania_analysis_upstreams__'], False)
    catalog.append(source)

    print(f'Chart is served as {chart_name} {chart_version}')


def validation_info_text(info, nest: int, name):
    t = '\t' * nest
    texts = [f'{t}- {name} | {info[name][-1][1]}']

    for upname, upflag in info[name][:-1]:
        texts.append(validation_info_text(info, nest + 1, upname))
    return os.linesep.join(texts)


@register_line_magic
def validate_catalog(line):
    catalog = FramaniaExtendedIntakeCatalog(get_ipython().user_global_ns['__framania_catalog_file__'])
    flag, info = catalog.validate()
    print(f'Validation {flag}')
    for name in sorted(info.keys()):
        print(validation_info_text(info, 0, name))



from typing import List

import pandas
from pandas import Series


def str_startswith_many(ps: Series, startswiths: List[str]):
    """
        API to check string column values start with some of specified string sequences.

        Args:
            ps (pandas.Series): target pandas dataframe
            startswiths (List[str]): string sequences to check string column startswith
        Returns:
            bool pandas series
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a': ['abc', 'bcd', 'cde', 'def']})
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                 a
            0  abc
            1  bcd
            2  cde
            3  def
            >>> result = str_startswith_many(pd.a, ['ab', 'bc'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
            0     True
            1     True
            2    False
            3    False
            dtype: bool
        """
    result = pandas.Series(False, index=ps.index)
    for swith in startswiths:
        result |= ps.str.startswith(swith)
    return result


def str_endswith_many(ps: Series, endswiths: List[str]):
    """
        API to check string column values end with some of specified string sequences.

        Args:
            ps (pandas.Series): target pandas dataframe
            endswiths (List[str]): string sequences to check string column endswith
        Returns:
            bool pandas series
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a': ['abc', 'bcd', 'cde', 'def']})
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                 a
            0  abc
            1  bcd
            2  cde
            3  def
            >>> result = str_endswith_many(pd.a, ['de', 'ef'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
            0    False
            1    False
            2     True
            3     True
            dtype: bool
        """
    result = pandas.Series(False, index=ps.index)
    for ewith in endswiths:
        result |= ps.str.endswith(ewith)
    return result


def str_contains_many(ps: Series, contains: List[str]):
    """
        API to check string column values contain some of specified string sequences.

        Args:
            ps (pandas.Series): target pandas dataframe
            contains (List[str]): string sequences to check string column contains
        Returns:
            bool pandas series
        Examples:
            >>> import pandas
            >>> pd = pandas.DataFrame({'a': ['abc', 'bcd', 'cde', 'def']})
            >>> print(pd)
            ... # doctest: +NORMALIZE_WHITESPACE
                 a
            0  abc
            1  bcd
            2  cde
            3  def
            >>> result = str_contains_many(pd.a, ['abc', 'def'])
            >>> print(result)
            ... # doctest: +NORMALIZE_WHITESPACE
            0     True
            1    False
            2    False
            3     True
            dtype: bool
        """
    result = pandas.Series(False, index=ps.index)
    for c in contains:
        result |= ps.str.contains(c)
    return result



from pathlib import Path
from typing import Union
from urllib.parse import urlparse
from uuid import uuid4

import dask
import pandas
import yaml
from dask.dataframe import DataFrame
from intake import DataSource
from intake_parquet import ParquetSource


class S3URL(str):
    """
    pathlib.Path like s3 url wrapper class.
    """

    def __new__(cls, content: str):
        if not content.startswith('s3://'):
            content = f's3://{content}'
        return str.__new__(cls, content)

    def __truediv__(self, other):
        if not self.endswith('/'):
            return S3URL(f'{self}/{other}')
        else:
            return S3URL(f'{self}{other}')

    def bucket(self):
        parsed = urlparse(self)
        return parsed.netloc

    def key(self):
        parsed = urlparse(self)
        return parsed.path.lstrip('/')

    def mkdir(self, **kwargs):
        pass


def local_or_s3_path(v: Union[str, Path, S3URL]) -> Union[Path, S3URL]:
    """
    Convert input value into local path(pathlib.Path) or S3URL.
    When yes, then return input value.
    When no and value starts with s3://, then return s3 remote url.
    Other case return local path

    Args:
        v: str, local path, or s3 remote url.
    Returns:
        local path or s3 remote url
    Examples:
        >>> assert isinstance(local_or_s3_path(Path('abc')), Path)
        >>> assert isinstance(local_or_s3_path(S3URL('abc')), S3URL)
        >>> assert isinstance(local_or_s3_path('abc'), Path)
        >>> assert isinstance(local_or_s3_path('s3://abc'), S3URL)
    """
    if isinstance(v, Path):
        return v
    if isinstance(v, S3URL):
        return v
    if isinstance(v, str) and v.startswith('s3://'):
        return S3URL(v)
    else:
        return Path(v)


def initialize_catalog(catalog_file: Union[Path, S3URL, str]):
    catalog_file = local_or_s3_path(catalog_file)
    try:
        catalog_contents = yaml.load(catalog_file.open().read(), Loader=yaml.SafeLoader)
    except FileNotFoundError:
        catalog_contents = {'sources': {}}
    if catalog_contents is None or \
            not isinstance(catalog_contents['sources'], dict) or \
            'sources' not in catalog_contents:
        catalog_contents = {'sources': {}}

    catalog_file.open('wt').write(yaml.dump(catalog_contents, Dumper=yaml.SafeDumper))


def add_source_to_catalog(source: DataSource, catalog_file: Union[Path, str]):
    """
    API to add new data source to catalog_file.

    Args:
        source: data source to add.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
    Examples:
        >>> import os
        >>> from intake.source.csv import CSVSource
        >>> import yaml
        >>> source1 = CSVSource('test/temp/test1.csv')
        >>> source1.name = 'csv-test1'
        >>> # create new catalog with source
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> add_source_to_catalog(source1, cfile)
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'csv-test1': {'args': {'urlpath': 'test/temp/test1.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}}}
        >>> source2 = CSVSource('test/temp/test2.csv')
        >>> source2.name = 'csv-test2'
        >>> add_source_to_catalog(source2, cfile)
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'csv-test1': {'args': {'urlpath': 'test/temp/test1.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}, 'csv-test2': {'args': {'urlpath': 'test/temp/test2.csv'}, 'description': '', 'driver': 'intake.source.csv.CSVSource', 'metadata': {}}}}
        >>> os.remove(cfile)
    """
    catalog_file = local_or_s3_path(catalog_file)
    try:
        catalog_contents = yaml.load(catalog_file.open().read(), Loader=yaml.SafeLoader)
    except FileNotFoundError:
        catalog_contents = {'sources': {}}
    if catalog_contents is None or \
            not isinstance(catalog_contents['sources'], dict) or \
            'sources' not in catalog_contents:
        catalog_contents = {'sources': {}}
    catalog_contents['sources'][source.name] = source._yaml()['sources'][source.name]

    catalog_file.open('wt').write(yaml.dump(catalog_contents, Dumper=yaml.SafeDumper))


def dump_dask_to_intake(dd: DataFrame, data_name: str, data_dir: Union[str, Path], catalog_file: Union[Path, str],
                        **kwargs):
    """
    API to dump dask dataframe as parquet format and add it to intake catalog

    Args:
        dd: dask dataframe to dump.
        data_name: name used as a name of intake data source.
        data_dir: directory where dask dataframe will be stored.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> print(dd)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask DataFrame Structure:
                           a      b      c   label
        npartitions=2
        100            int64  int64  int64  object
        300              ...    ...    ...     ...
        400              ...    ...    ...     ...
        Dask Name: from_pandas, 2 tasks
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> ddir = 'test/temp/data-dir'
        >>> # DUMP WITH COMPUTATION
        >>> psource1, job = dump_dask_to_intake(dd, 'test-dd1', ddir, cfile)
        >>> print(psource1.name)
        test-dd1
        >>> print(psource1.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> print(job is None)
        True
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'test-dd1': {'args': {'urlpath': 'test/temp/data-dir/test-dd1'}, 'description': '', 'driver': 'intake_parquet.source.ParquetSource', 'metadata': {}}}}
        >>> # DUMP WITHOUT COMPUTATION
        >>> psource2, job = dump_dask_to_intake(dd, 'test-dd2', ddir, cfile, compute=False)
        >>> print(job is None)
        False
        >>> # do computation lazily
        >>> job.compute()
        >>> print(psource2.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> os.remove(cfile)
        >>> shutil.rmtree(ddir)
    """
    data_dir = local_or_s3_path(data_dir)

    parquet_dir = data_dir / data_name

    parquet_kwargs = {}
    if 'engine' in kwargs:
        parquet_kwargs['engine'] = kwargs['engine']

    psource = ParquetSource(str(parquet_dir), **parquet_kwargs)
    psource.name = data_name
    add_source_to_catalog(psource, catalog_file)
    to_parquet_result = dd.to_parquet(str(parquet_dir), **kwargs)

    return psource, to_parquet_result


def dump_pandas_to_intake(pd: pandas.DataFrame, data_name: str, data_dir: Union[str, Path],
                          catalog_file: Union[Path, str], npartitions: int = 1, **kwargs):
    """
    API to dump pandas dataframe as parquet format and add it to intake catalog.
    pandas dataframe will be converted to dask dataframe automatically internally.

    Args:
        pd: pandas dataframe to dump.
        data_name: name used as a name of intake data source.
        data_dir: directory where dask dataframe will be stored.
        catalog_file: file where data source to be added. if file doesn't exist, file will be created.
        npartitions: number of partitions input pandas dataframe split into.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> cfile = 'test/temp/test-catalog.yaml'
        >>> ddir = 'test/temp/data-dir'
        >>> # DUMP WITH COMPUTATION
        >>> psource1, job = dump_pandas_to_intake(pd, 'test-pd1', ddir, cfile)
        >>> print(psource1.name)
        test-pd1
        >>> print(psource1.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> print(job is None)
        True
        >>> print(yaml.safe_load(Path(cfile).open().read()))
        ... # doctest: +NORMALIZE_WHITESPACE
        {'sources': {'test-pd1': {'args': {'urlpath': 'test/temp/data-dir/test-pd1'}, 'description': '', 'driver': 'intake_parquet.source.ParquetSource', 'metadata': {}}}}
        >>> # DUMP WITHOUT COMPUTATION
        >>> psource2, job = dump_pandas_to_intake(pd, 'test-pd2', ddir, cfile, compute=False)
        >>> print(job is None)
        False
        >>> # do computation lazily
        >>> job.compute()
        >>> print(psource2.read())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> os.remove(cfile)
        >>> shutil.rmtree(ddir)
    """
    dd = dask.dataframe.from_pandas(pd, npartitions=npartitions)
    return dump_dask_to_intake(dd, data_name, data_dir, catalog_file, **kwargs)


def persist_local(dd: DataFrame, persist_dir: Union[str, Path], **kwargs):
    """
    API to persist dask dataframe as perquet files on disk.

    Args:
        dd: dask dataframe to persist.
        persist_dir: directory where dask dataframe will be stored.
        kwargs: Any options available for dask.dataframe.to_parquet. see https://docs.dask.org/en/latest/dataframe-api.html#dask.dataframe.to_parquet for detail.
    Returns:
        created parquet data source and dask to_parquet job (if you put compute=False in kwargs.)
    Examples:
        >>> import os
        >>> import shutil
        >>> import yaml
        >>> from intake.source.csv import CSVSource
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 3, 4, 5], 'c': [5, 6, 7, 8],
        ...                        'label': ['a', 'b', 'c', 'd']},
        ...                        index=[100, 200, 300, 400])
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
             a  b  c label
        100  1  2  5     a
        200  2  3  6     b
        300  3  4  7     c
        400  4  5  8     d
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> ddir = 'test/temp/data-dir/persist'
        >>> dd2 = persist_local(dd, ddir)
        >>> persist_id = dd2.persist_id
        >>> print(dd2.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b  c label
        index
        100    1  2  5     a
        200    2  3  6     b
        300    3  4  7     c
        400    4  5  8     d
        >>> assert (Path(ddir) / persist_id).exists()
        >>> shutil.rmtree(ddir)
    """
    data_id = str(uuid4())
    persist_dir = local_or_s3_path(persist_dir)

    persist_dir.mkdir(parents=True, exist_ok=True)

    persist_catalog_file = Path(persist_dir) / 'persist_catalog.yaml'
    psource, _ = dump_dask_to_intake(dd, data_id, persist_dir, persist_catalog_file, **kwargs)

    new_dd = psource.to_dask()
    new_dd.persist_id = data_id

    return new_dd



from tempfile import mkdtemp, mktemp
from unittest import TestCase

import intake
import pandas
from dask.dataframe.utils import make_meta
from intake.catalog.local import YAMLFileCatalog
from intake.source.csv import CSVSource

from framania.intakemania.user_defined_class_parquet_source import upload_with_user_defined_class, UserDefinedClassParquetSource
from framania.pandasmania.util import md5hash


class MyClass:
    def __init__(self, x):
        self._x = x

    def __eq__(self, other):
        return self._x == other._x

    def __hash__(self):
        return hash(self._x)


class TestFramaniaExtendedIntake(TestCase):
    def setUp(self) -> None:
        self.csvsource1 = CSVSource(urlpath='test/data/test-csv1.csv')

    @staticmethod
    def create_custom_source(df):
        filename = mktemp()
        df["d"] = df["b"].map(lambda x: MyClass(x), meta=make_meta(df["b"]))
        source = upload_with_user_defined_class(df, filename, ["d"])

        return source, filename

    def test_upload_with_custom_class(self):
        df = self.csvsource1.to_dask()
        source, _ = self.create_custom_source(df)
        assert source.metadata["user_defined_class_columns"] == ["d"]
        assert df["d"].head().iloc[0] == MyClass("a")
        assert isinstance(source, UserDefinedClassParquetSource)

    def test_load_with_custom_class(self):
        df = self.csvsource1.to_dask()
        source, _ = self.create_custom_source(df)
        yaml_file = mktemp(suffix=".yaml")
        f = open(yaml_file, "w")
        f.write("""\
metadata:
  version: 1
sources: {}""")
        f.close()
        cat = intake.open_catalog(yaml_file)
        cat.add(source, name="test")

        df2 = source.to_dask()

        assert all(df["d"].compute() == df2["d"].compute())
        assert all(df.compute() == df2.compute())
        # TODO(higumachan): これが通らない
        # assert md5hash(df.compute()) == md5hash(df2.compute())

    def test_persist_user_defined_parquet_source(self):
        df = self.csvsource1.to_dask()
        source, filename = self.create_custom_source(df)

        source.persist()

        assert source.has_been_persisted



import hashlib
from typing import Union, Dict, Any, Optional

import pandas
from dask.dataframe import DataFrame, Series

from framania.pandasmania.util import md5hash as md5hash_pandas


def dataframe_from_series_of_pandas(series_of_pandas_dataframes: Series,
                                    schema: Optional[Union[pandas.DataFrame, DataFrame, Dict[Any, Any]]] = None):
    """
    pandas.DataFrameが各行に保持されているdask.dataframe.Seriesをdask.dataframe.Dataframeに変換するAPI。
    schemaを指定すればそれを結果のスキーマとして使用し、Noneなら第１パーティションの第１行に保持されているdataframeを使用する。

    Args:
        series_of_pandas_dataframes: dask dataframe from series of pandas dataframes.
        schema: schema of resutl dataframe
    Return: dask dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> df = pandas.DataFrame({'a': range(100), 'b': range(100, 200)})
        >>> print(df)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b
        0    0  100
        1    1  101
        2    2  102
        3    3  103
        4    4  104
        ..  ..  ...
        95  95  195
        96  96  196
        97  97  197
        98  98  198
        99  99  199
        [100 rows x 2 columns]
        >>> ddf = dask.dataframe.from_pandas(df, npartitions=4)
        >>> def build_df_from_row(row):
        ...     pdf = pandas.DataFrame({'values': [row.a, row.a + 1, row.a + 2], 'support': [row.b, row.b, row.b]})
        ...     return pdf
        >>> df_ds = ddf.apply(build_df_from_row, axis=1)
        >>> print(df_ds)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask Series Structure:
        npartitions=4
        0     object
        25       ...
        50       ...
        75       ...
        99       ...
        dtype: object
        Dask Name: apply, 8 tasks
        >>> result = dataframe_from_series_of_pandas(df_ds)
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask DataFrame Structure:
                      values support
        npartitions=4
        0              int64   int64
        25               ...     ...
        50               ...     ...
        75               ...     ...
        99               ...     ...
        Dask Name: create_pandas_dataframe_in_partition, 12 tasks
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
                values  support
        0        0      100
        1        1      100
        2        2      100
        0        1      101
        1        2      101
        ..     ...      ...
        1       99      198
        2      100      198
        0       99      199
        1      100      199
        2      101      199
        [300 rows x 2 columns]
    """
    if schema is None:
        schema = series_of_pandas_dataframes.head(1).iloc[0]

    def create_pandas_dataframe_in_partition(series_chunk: pandas.Series):
        df = pandas.concat(list(series_chunk), axis=0)
        return df

    ddf = series_of_pandas_dataframes.map_partitions(create_pandas_dataframe_in_partition, meta=schema)
    return ddf


def dataframe_from_series_of_record_dict(series_of_record_dicts: Series,
                                         schema: Union[pandas.DataFrame, DataFrame, Dict[Any, Any]]):
    """
    Dictが各行に保持されているdask.dataframe.Seriesをdask.dataframe.Dataframeに変換するAPI。
    schemaを指定すればそれを結果のスキーマとして使用し、Noneなら第１パーティションの第１行に保持されているdataframeを使用する。

    Args:
        series_of_record_dicts: dask dataframe from series of pandas dataframes.
        schema: schema of resutl dataframe
    Return: dask dataframe
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> df = pandas.DataFrame({'a': range(100), 'b': range(100, 200)})
        >>> print(df)
        ... # doctest: +NORMALIZE_WHITESPACE
             a    b
        0    0  100
        1    1  101
        2    2  102
        3    3  103
        4    4  104
        ..  ..  ...
        95  95  195
        96  96  196
        97  97  197
        98  98  198
        99  99  199
        [100 rows x 2 columns]
        >>> ddf = dask.dataframe.from_pandas(df, npartitions=4)
        >>> def build_df_from_row(row):
        ...     return [{'values': row.a, 'support': row.b},
        ...             {'values': row.a + 1, 'support': row.b},
        ...             {'values': row.a + 2, 'support': row.b}]
        >>> df_ds = ddf.apply(build_df_from_row, axis=1, meta='object')
        >>> print(df_ds)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask Series Structure:
        npartitions=4
        0     object
        25       ...
        50       ...
        75       ...
        99       ...
        dtype: object
        Dask Name: apply, 8 tasks
        >>> result = dataframe_from_series_of_record_dict(df_ds, {'values': 'int', 'support': 'int'})
        >>> print(result)
        ... # doctest: +NORMALIZE_WHITESPACE
        Dask DataFrame Structure:
                      values support
        npartitions=4
        0              int64   int64
        25               ...     ...
        50               ...     ...
        75               ...     ...
        99               ...     ...
        Dask Name: create_pandas_dataframe_in_partition, 12 tasks
        >>> print(result.compute())
        ... # doctest: +NORMALIZE_WHITESPACE
                values  support
        0        0      100
        1        1      100
        2        2      100
        0        1      101
        1        2      101
        ..     ...      ...
        1       99      198
        2      100      198
        0       99      199
        1      100      199
        2      101      199
        [300 rows x 2 columns]
    """
    def create_pandas_dataframe_in_partition(series_chunk: pandas.Series):
        df = pandas.concat(list(series_chunk.apply(pandas.DataFrame.from_records)), axis=0)
        return df

    ddf = series_of_record_dicts.map_partitions(create_pandas_dataframe_in_partition, meta=schema)
    return ddf


def md5hash(dd: DataFrame) -> hashlib.md5:
    """
    API to make md5hash from pandas dataframe.

    Args:
        pd: pandas dataframe.
    Examples:
        >>> import dask.dataframe
        >>> import pandas
        >>> pd = pandas.DataFrame({'a': [1, 10, 100, 1, 1, 100], 'b': ['a', 'b', 'c', 'd', 'e', 'f'],
        ...                       'idx': [0, 1, 2, 0, 1, 2]}).set_index('idx')
        >>> print(pd)
        ... # doctest: +NORMALIZE_WHITESPACE
               a  b
        idx
        0      1  a
        1     10  b
        2    100  c
        0      1  d
        1      1  e
        2    100  f
        >>> dd = dask.dataframe.from_pandas(pd, npartitions=2)
        >>> md5hash(dd)
        'e650778d6cdca953fad30cccedd4c5f1'
    """
    results = dd.map_partitions(md5hash_pandas).compute()
    m = hashlib.md5()
    for v in results:
        m.update(v.encode('utf-8'))
    return m.hexdigest()
